#!/usr/bin/perl
#
#	hostAttackTree.pl
#
#	Reads adjacency files and makes a host-account-host tree
#	graph.
#
#	This code is by Foofus (foofus at foofus dot net)
#

# Initialize default values
$version="0,1";
$mapdir="./maps";
$adjdir="./output/adjacency";
$dotdir="./output/dot";
$prefix="map";
$colorMapFile="colors.map";
$shapeMapFile="shapes.map";
$outfile="hostAttackTree.dot";
$targetHost="";

# Process command-line arguments
foreach $arg (@ARGV)
{
	chomp($arg);
	@pair=split(/:/,$arg);
	if (@pair[0] eq "in")
	{
		$indir=@pair[1];
	} elsif (@pair[0] eq "ad") {
		$adjdir=@pair[1];
	} elsif (@pair[0] eq "th") {
		$targetHost=@pair[1];
	} elsif (@pair[0] eq "d2") {
		$domainTwo=@pair[1];
	} elsif (@pair[0] eq "pr") {
		$prefix=@pair[1];
	} elsif (@pair[0] eq "el") {
	        $exfile=@pair[1];
	} elsif (@pair[0] eq "sm") {
		$shapeMapFile=@pair[1];
	} elsif (@pair[0] eq "cm") {
		$colorMapFile=@pair[1];
	} else {
		print "@pair[0] is not a supported option.\n";
		&showUsage;
		die;
	}
} # (foreach)

# Sanity check goes here
print "Welcome to hostAttackTree.pl version $version by Foofus.\n";
print "Adjacency files are in $adjdir\n";
print "Map files are in $mapdir\n";
print "Color map is $colorMapFile\n";
print "Shape map is $shapeMapFile\n";
print "These files will be identified by the prefix \"$prefix\"\n";
if ($targetHost eq ""){
	print "ERROR: You must specify a domain with the th: parameter!\n";
	print "       Otherwise, there's just no point.\n";
	die;
} else {
	print "Building a tree for $targetHost\n";
}

# Build color map
%colorMap=("NODE_TYPE" => "NODE_COLOR");
open(INFILE,"<$mapdir/$colorMapFile") || die "Error: Unable to open $mapdir/$colorMapFile for reading!\n";
@rawColors=<INFILE>;
close(INFILE);
foreach $entry (@rawColors)
{
	chomp($entry);
	@pair=split(/:/,$entry);
	$colorMap{@pair[0]}=@pair[1];
} # (foreach)

# Build shape map
%shapeMap=("NODE_TYPE" => "NODE_SHAPE");
open(INFILE,"<$mapdir/$shapeMapFile") || die "Error: Unable to open $mapdir/$shapeMapFile for reading!\n";
@rawShapes=<INFILE>;
close(INFILE);
foreach $entry (@rawShapes)
{
	chomp($entry);
	@pair=split(/:/,$entry);
	$shapeMap{@pair[0]}=@pair[1];
} # (foreach)

# Build the list of hosts by host
open(INFILE,"<$adjdir/$prefix-HostsByHost.adj") || die "ERROR: Unable to open $adjdir/$prefix-HostsByHost.adj for reading!\n";
@rawHostsByHost=<INFILE>;
close(INFILE);
# Build the adjacency hash
%hostsByHost=("HostName" => "Neighbors");
foreach $entry (@rawHostsByHost)
{
	chomp($entry);
	@pair=split(/:/,$entry);
	$thingy=@pair[0];
	$hostsByHost{$thingy}=@pair[1];
} # (foreach)

# Build the list of accounts by host
open(INFILE,"<$adjdir/$prefix-AccountsByHost.adj") || die "ERROR: Unable to open $adjdir/$prefix-AccountsByHost.adj for reading!\n";
@rawAccountsByHost=<INFILE>;
close(INFILE);
# Build the adjacency hash
%accountsByHost=("HostName" => "Accounts");
foreach $entry (@rawAccountsByHost)
{
	chomp($entry);
	@pair=split(/:/,$entry);
	$accountsByHost{@pair[0]}=@pair[1];
} # (foreach)

# Build the list of hosts by account
open(INFILE,"<$adjdir/$prefix-HostsByAccount.adj") || die "ERROR: Unable to open $adjdir/$prefix-HostsByAccount.adj for reading!\n";
@rawHostsByAccount=<INFILE>;
close(INFILE);
# Build the adjacency hash
%hostsByAccount=("AccountName" => "Hosts");
foreach $entry (@rawHostsByAccount)
{
	chomp($entry);
	@pair=split(/:/,$entry);
	$hostsByAccount{"@pair[0]"}=@pair[1];
} # (foreach)

# Begin the output file
open(OUTFILE,">$dotdir/$prefix-$outfile") || die "ERROT: Unable to open $dotdir/$prefix-$outfile for writing!\n";
print OUTFILE "/* This dot file was generated by hostAttackTree.pl version $version by Foofus */\n";
print OUTFILE "/* Output is based on $adjdir/$prefix-HostsByHost.adj, $adjdir/$prefix-HostsByAccount.adj, and $adjdir/$prefix-AccountsByHost.adj */\n";
print OUTFILE "/* Color map : $mapdir/$colorMapFile */\n";
print OUTFILE "/* Shape map : $mapdir/$shapeMapFile */\n";
print OUTFILE "/* Target host is $targetHost */\n";
print OUTFILE "Graph G {\n";

%hostsDrawn=("HostName" => "Drawn");
%accountsDrawn=("Account" => "Drawn");
$edgesDrawn="none";

# Draw the head node
$shape=$shapeMap{"HOST"};
$color=$colorMap{"HOST"};
print OUTFILE "  /* Root node is $targetHost */\n";
print OUTFILE "  \"$targetHost\" [label=\"$targetHost\",fontname=\"Courier\",shape=\"$shape\",color=\"$color\",style=filled];\n";
$hostsDrawn{$targetHost}="Drawn";

# Get the list of adjacent hosts
$nextRank=$hostsByHost{$targetHost};
$previousRank=$targetHost;

# While there are children
while (&hasChildren($nextRank)>0)
{
	# Define a subgraph for this rank
	print OUTFILE "    { rank=same;\n";
	#print OUTFILE "    { \n";
	# Draw the hosts in this rank
	@rankList=split(/,/,$nextRank);
	foreach $host (@rankList)
	{
		if (exists($hostsDrawn{$host}))
		{
			print OUTFILE "      /* $host has already been drawn */\n";
		} else {
			$shape=$shapeMap{"HOST"};
			$color=$colorMap{"HOST"};
			print OUTFILE "      \"$host\" [label=\"$host\",fontname=\"Courier\",shape=\"$shape\",color=\"$color\",style=filled];\n";
			$hostsDrawn{$host}="Drawn";
		} # (has this host already been drawn?)
	} # (foreach)
	# Terminate the subgraph
	print OUTFILE "    }\n";

	# Now draw edges
	@parentList=split(/,/,$previousRank);
	foreach $parent (@parentList)
	{
		foreach $child (@rankList)
		{
			if ($child ne $parent)
			{
			chomp($parent);
			chomp($child);
			if ((index($edgesDrawn,"$parent-$child")>0)||(index($edgesDrawn,"$child-$parent")>0))
			{
				print OUTFILE "    /* We already had a $parent -> $child edge */\n";
			} else {
				if (exists($hostsByHost{$parent}))
				{
					$pc=$hostsByHost{$parent};
					if (index($pc,$child)>-1)
					{
						print OUTFILE "  \"$parent\" -- \"$child\";\n";
						$edgesDrawn=$edgesDrawn."$parent-$child";
					}
				}
			}
			}
		} # (foreach)
	} # (foreach)
	# Set up for next pass (nextRank, previousRank)
	$previousRank=$nextRank;
	$nextRank="";
	foreach $host (@rankList)
	{
		$rawNext=$hostsByHost{$host};
		print "$host --- $rawNext\n";
		@nextList=split(/,/,$rawNext);
		foreach $possible (@nextList)
		{
			if ((index($nextRank,$possible) == -1)&&(!(exists($hostDrawn{$possible}))))
			{
				if ($nextRank ne "")
				{
					$nextRank=$nextRank.",";
				}
				$nextRank=$nextRank.$possible;
			}
		} # (foreach)
	} # (foreach)
	print "nextRank : $nextRank\n";
} # (while)

print OUTFILE "}\n";
close(OUTFILE);

###
### Subroutines begin here, now that the main program is all done
###

#	hasChildren -- calculates the number of hosts in a packed rank that have children
sub hasChildren {
	local($packedRank)=@_;
	my(@rankList,$retVal);

	chomp($packedRank);
	$retVal=0;
	@rankList=split(/,/,$packedRank);

	foreach $node (@rankList)
	{
		if (exists($hostsByHost{$node}))
		{
			if (exists($hostsDrawn{$node}))
			{ 
			#	print "Wha? $node\n";
			} else {
				$retVal++;
			}
		}
	} # (foreach)

	return $retVal;
} # (hasChildren)

#	drawHost -- draws all accounts descending from a given host
sub drawHost {
	local($host)=@_;
	my($shape,$color,$account,$accounts,@accountList);
	chomp($host);

	# Have we drawn this host before?
	if (exists($hostsDrawn{$host}))
	{
		print OUTFILE "    /* $host is already represented above */\n";
	} else {
	
	$hostsDrawn{$host}="Drawn";

	$shape=$shapeMap{"HOST"};
	$color=$colorMap{"HOST"};
	print OUTFILE "    \"$host\" [label=\"$host\",fontname=\"Courier\",shape=\"$shape\",color=\"$color\",style=filled];\n";

	if (exists($accountsByHost{$host}))
	{
		# Make a subgraph
		#print OUTFILE "    /* Subgraph for accounts descending from $host */\n";
		#print OUTFILE "    { \n      rank=source;\n";
		#print OUTFILE "    { \n";

		$accounts=$accountsByHost{$host};
		@accountList=split(/,/,$accounts);
		foreach $account (@accountList)
		{
			&drawAccount($account);
		} # (foreach)
		# End the subgraph 
		#print OUTFILE "    } \n";
		# Draw edges 
		foreach $account (@accountList)
		{
			print OUTFILE "    \"$host\" -- \"$account\";\n";
		} # (foreach)
	} else {
		print OUTFILE "    /* Host $host has no descendant accounts */\n";
	}
	} # (already drawn?)
} # (drawHost)

sub drawAccount {
	local($account)=@_;
	chomp($account);
	my($hosts,$shape,$color,$host,$hostList);

	# Have we drawn this account before?
	if (exists($accountsDrawn{$account}))
	{
		print OUTFILE "    /* $account is already represented above */\n";
	} else {

	$accountsDrawn{$account}="Drawn";

	$shape=$shapeMap{"ACCOUNT"};
	$color=$colorMap{"ACCOUNT"};
	print OUTFILE "    \"$account\" [label=\"$account\",fontname=\"Nimbus\",shape=\"$shape\",color=\"$color\",style=filled];\n";

	# Make a subgraph
	#print OUTFILE "    /* Subgraph for hosts descending from $account */\n";
	#print OUTFILE "    { \n      rank=source;\n";
	#print OUTFILE "    { \n";
	$hosts=$hostsByAccount{$account};
	@hostList=split(/,/,$hosts);
	foreach $host (@hostList)
	{
		&drawHost($host);
	} # (foreach)
	# End the subgraph
	#print OUTFILE "    } \n";
	foreach $host (@hostlist)
	{
		print OUTFILE "    \"$account\" -- \"$host\";\n";
	} # (foreach)
	} # (already drawn?)
} # (drawAccount)

#	showUsage -- shows, um, usage
sub showUsage {
	print "mapAdjacencies.pl -- version $version by Foofus\n";
	print "This program builds lists of adjacencies in Windows password dump\n";
	print "files, and gathers some statistics.\n";
	print "usage: mapAdgacencies.pl [parameter:value] [parameter:value] [...]\n";
	print "Available parameters\n";
	print "th : the name of the target host (i.e., the root of the tree)\n";
	print "     THIS PARAMETER IS REQUIRED!\n";
	print "ad : the directory where adjacency files are located\n";
	print "     (default is $adjdir)\n";
	print "sf : suffix for pwdump/fgdump files (default is $suffix)\n";
	print "pr : prefix to prepend to output files (default is \"$prefix\")\n";
	print "el : file where account exclusion strings reside\n";
	print "     (default is $exfile) \n";
}	# (showUsage)

